<?php

declare(strict_types=1);

namespace Drupal\Tests\graphql_webform\Kernel\Mutation;

use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Tests\graphql_webform\Kernel\GraphQLWebformKernelTestBase;
use Drupal\webform\WebformEntityStorageInterface;
use Drupal\webform\WebformInterface;
use Drupal\webform\WebformSubmissionStorageInterface;

/**
 * Tests submission of a webform.
 *
 * @group graphql_webform
 */
class FormSubmissionTest extends GraphQLWebformKernelTestBase {

  /**
   * The entity type manager.
   */
  protected ?EntityTypeManagerInterface $entityTypeManager;

  /**
   * The Webform storage.
   */
  protected ?WebformEntityStorageInterface $webformStorage;

  /**
   * The Webform submission storage.
   */
  protected ?WebformSubmissionStorageInterface $webformSubmissionStorage;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    // The webform module requires the File module when a form allows file
    // uploads.
    'file',
    // The webform module depends silently on the Path Alias module, which is
    // fine since this is a required core module.
    'path_alias',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('user');
    $this->installEntitySchema('webform_submission');
    $this->installConfig('webform');

    $this->entityTypeManager = $this->container->get('entity_type.manager');
    $this->webformStorage = $this->entityTypeManager->getStorage('webform');
    $this->webformSubmissionStorage = $this->entityTypeManager->getStorage('webform_submission');
  }

  /**
   * Tests submitting a form with an invalid form ID.
   */
  public function testSubmitInvalidFormId(): void {
    $query = $this->getQueryFromFile('invalid_submission.gql');
    $variables = ['id' => 'invalid_form_id', 'elements' => []];
    $this->assertResults(
      $query,
      $variables,
      [
        'submitWebform' => [
          'errors' => ['Webform invalid_form_id does not exist.'],
          'validationErrors' => [],
          'submission' => NULL,
        ],
      ],
      // Since we do not have a valid webform to submit, only the basic cache
      // metadata is associated with the response.
      $this->baseCacheMetadata()
    );

    $this->assertWebformSubmissionCount(0, 'An invalid submission is not saved.');
  }

  /**
   * Tests that a form submission is rejected if the user does not have access.
   *
   * @param bool $authenticated_access_allowed
   *   Whether authenticated users are allowed to submit the form.
   *
   * @dataProvider submitAccessProvider
   */
  public function testSubmitAccess(bool $authenticated_access_allowed): void {
    if (!$authenticated_access_allowed) {
      // Update the form configuration to disallow authenticated access.
      $webform = $this->webformStorage->load('graphql_webform_test_form');
      $access_rules = $webform->getAccessRules();
      $access_rules['create']['roles'] = ['anonymous'];
      $webform->setAccessRules($access_rules)->save();
    }

    $query = $this->getQueryFromFile('invalid_submission.gql');
    $variables = [
      'id' => 'graphql_webform_test_form',
      'elements' => [(object) ['element' => 'required_text_field', 'value' => 'A value.']],
    ];

    $expected_cache_metadata = $this->baseCacheMetadata()
      ->addCacheContexts($this->defaultCacheContexts())
      ->setCacheTags($this->defaultCacheTags());
    // Add the submission cache tags if the submission was successful.
    if ($authenticated_access_allowed) {
      $expected_cache_metadata->addCacheTags(['webform_submission:1']);
    }

    $expected_errors = $authenticated_access_allowed ? [] : ['Please login to access this form.'];

    $expected_submission = $authenticated_access_allowed ? [
      'id' => '1',
      'webform' => [
        'id' => 'graphql_webform_test_form',
      ],
    ] : NULL;

    $this->assertResults($query, $variables, [
      'submitWebform' => [
        'errors' => $expected_errors,
        'validationErrors' => [],
        'submission' => $expected_submission,
      ],
    ], $expected_cache_metadata);

    $this->assertWebformSubmissionCount((int) $authenticated_access_allowed);
  }

  /**
   * Provides data for the testSubmitAccess test.
   *
   * @return array
   *   An array of test cases.
   */
  public function submitAccessProvider(): array {
    return [
      'authenticated access allowed' => [TRUE],
      'authenticated access not allowed' => [FALSE],
    ];
  }

  /**
   * Tests submitting a form with an invalid value.
   */
  public function testValidationError(): void {
    $query = $this->getQueryFromFile('invalid_submission.gql');
    $variables = [
      'id' => 'graphql_webform_test_form',
      'elements' => [(object) ['element' => 'required_text_field', 'value' => 'This value is too long.']],
    ];

    $expected_cache_metadata = $this->baseCacheMetadata()
      ->addCacheContexts($this->defaultCacheContexts())
      ->setCacheTags($this->defaultCacheTags());

    $this->assertResults($query, $variables, [
      'submitWebform' => [
        'errors' => [],
        'validationErrors' => [
          [
            'element' => 'required_text_field',
            'messages' => ['Required text field cannot be longer than 10 characters but is currently 23 characters long.'],
          ],
        ],
        'submission' => NULL,
      ],
    ], $expected_cache_metadata);

    $this->assertWebformSubmissionCount(0, 'An invalid submission is not saved.');
  }

  /**
   * Tests submitting a form that is closed.
   *
   * @param string $closed_message
   *   The closed message to set on the form.
   * @param string $default_closed_message
   *   The default closed message configured in the Webform global settings.
   * @param string $expected_error_message
   *   The expected error message.
   *
   * @dataProvider closedFormProvider
   */
  public function testClosedForm(string $closed_message, string $default_closed_message, string $expected_error_message): void {
    /** @var \Drupal\webform\WebformInterface $webform */
    $webform = $this->webformStorage->load('graphql_webform_test_form');
    $webform
      // Close the test form.
      ->setStatus(WebformInterface::STATUS_CLOSED)
      // Set the closed message.
      ->setSetting('form_close_message', $closed_message)
      ->save();

    // Set the default closed message.
    $this->config('webform.settings')
      ->set('settings.default_form_close_message', $default_closed_message)
      ->save();

    $query = $this->getQueryFromFile('invalid_submission.gql');
    $variables = [
      'id' => 'graphql_webform_test_form',
      'elements' => [(object) ['element' => 'required_text_field', 'value' => 'A value.']],
    ];

    $this->assertResults($query, $variables, [
      'submitWebform' => [
        'errors' => [$expected_error_message],
        'validationErrors' => [],
        'submission' => NULL,
      ],
    ], $this->baseCacheMetadata()
      ->addCacheContexts($this->defaultCacheContexts())
      ->setCacheTags($this->defaultCacheTags()));

    $this->assertWebformSubmissionCount(0, 'A submission for a closed Webform is not saved.');
  }

  /**
   * Provides data for the testClosedForm test.
   *
   * @return array
   *   An array of test cases.
   */
  public function closedFormProvider(): array {
    return [
      'custom message' => [
        '<p>This form is closed because of reasons.</p>',
        'Sorry… This form is closed to new submissions.',
        'This form is closed because of reasons.',
      ],
      'default message' => [
        '',
        'Sorry… This form is closed to new submissions.',
        'Sorry… This form is closed to new submissions.',
      ],
      'fallback message' => [
        '',
        '',
        'This form is closed to new submissions.',
      ],
    ];
  }

  /**
   * Tests submitting a form with a required field omitted.
   *
   * This should return the required text that was configured in the webform.
   */
  public function testOmitRequiredField(): void {
    $query = $this->getQueryFromFile('invalid_submission.gql');
    $variables = [
      'id' => 'graphql_webform_test_form',
      'elements' => [],
    ];

    $this->assertResults($query, $variables, [
      'submitWebform' => [
        'errors' => [],
        'validationErrors' => [
          [
            'element' => 'required_text_field',
            'messages' => ['This field is required because it is important.'],
          ],
        ],
        'submission' => NULL,
      ],
    ], $this->baseCacheMetadata()
      ->addCacheContexts($this->defaultCacheContexts())
      ->setCacheTags($this->defaultCacheTags())
    );

    $this->assertWebformSubmissionCount(0, 'An invalid submission is not saved.');
  }

  /**
   * Tests a valid form submission.
   */
  public function testValidSubmission(): void {
    $query = $this->getQueryFromFile('invalid_submission.gql');
    $variables = [
      'id' => 'graphql_webform_test_form',
      'elements' => [(object) ['element' => 'required_text_field', 'value' => 'A value.']],
    ];

    $this->assertResults($query, $variables, [
      'submitWebform' => [
        'errors' => [],
        'validationErrors' => [],
        'submission' => [
          'id' => '1',
          'webform' => [
            'id' => 'graphql_webform_test_form',
          ],
        ],
      ],
    ], $this->defaultCacheMetaData()->addCacheTags(['webform_submission:1']));

    // There should be one submission in the database.
    $this->assertWebformSubmissionCount(1, 'A valid submission is saved.');

    // Load the webform submission and check it contains the expected value.
    $submissions = $this->webformSubmissionStorage->loadMultiple();
    $submission = reset($submissions);
    $this->assertEquals('A value.', $submission->getElementData('required_text_field'));
  }

  /**
   * {@inheritdoc}
   */
  protected function defaultCacheContexts(): array {
    return ['user.permissions', 'languages:language_interface'];
  }

  /**
   * {@inheritdoc}
   */
  protected function defaultCacheMaxAge(): int {
    // Since these are mutations we don't want to cache the results. A
    // subsequent request will probably have different results.
    return 0;
  }

  /**
   * Returns the minimal cache metadata associated with any GraphQL response.
   *
   * @return \Drupal\Core\Cache\CacheableMetadata
   *   The cache metadata object.
   */
  protected function baseCacheMetadata(): CacheableMetadata {
    return (new CacheableMetadata())
      ->setCacheMaxAge($this->defaultCacheMaxAge())
      ->setCacheContexts(['user.permissions'])
      ->setCacheTags(['graphql_response', "config:graphql.graphql_servers.{$this->server->id()}"]);
  }

  /**
   * Checks the number of Webform submissions.
   *
   * @param int $expected_count
   *   The expected number of Webform submissions.
   * @param string $message
   *   The assertion message.
   */
  protected function assertWebformSubmissionCount(int $expected_count, string $message = ''): void {
    $count = $this->webformSubmissionStorage->getTotal();
    $this->assertEquals($expected_count, $count, $message);
  }

}
